
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errors.h>
#include <file.h>
#include <util.h>
#include <image.h>

#define TRUNCSTR "File may be truncated.\n"

static int garbage;
static long numgot, filesize;

void SetOutputType(char *, const FileFormat, Image *);
FileFormat ReadFileType (const char *);
int ReadFile (char *, const FileFormat, Image *);
int WriteFile (char *, Image *);
int WritePNM (FILE *, byte *, const int, const int, const int, 
              const int, char *);
int WriteWNM (FILE *, Image *, const int, const int, const int, 
              const int, int *, char *);
int LoadPNM (const char *, Image *);
static int loadpgm (FILE *, Image *, int);
static int loadppm (FILE *, Image *, int);
int LoadWNM (const char *, Image *);
static int loadwgm (FILE *, Image *);
static int loadwpm (FILE *, Image *);
static int getint (FILE *, Image *);
static int getbit (FILE *, Image *);
static int getshort (FILE *);

//------------------------------------------------
void 
SetOutputType (char *outType, const FileFormat inType, Image *image)
{
    // Set output file type
    if ( !strcmp (outType, "pgm") ) {
        image->outFileType = F_PBMRAW;
        image->colType = 1;
    } else if ( !strcmp (outType, "ppm") ) {
        image->outFileType = F_PBMRAW;
        image->colType = 0;
    } else if ( !strcmp (outType, "wgm") ) {
        image->outFileType = F_WTFWLT;
    } else if ( !strcmp (outType, "wpm") ) {
        image->outFileType = F_WTFWLT;
    } else {
        // Set outType depending on inType
        switch (inType) {
            case FMT_PNM:
                image->outFileType = F_PBMRAW;
                break;
            case FMT_WNM:
                image->outFileType = F_WTFWLT;
                break;
            default:
                Error ("SetOutputType", NO_FORMAT, ABORT);
        }
    }
}

//------------------------------------------------
FileFormat 
ReadFileType ( const char *filename )
{
    FILE       *fp;
    byte        magicno[10];
    FileFormat  type;
    unsigned int n;

    if ( !filename )    
        return FMT_ERROR;

    if ( (fp = fopen (filename, "r")) == NULL ) {
        Error ("OpenFile", NO_FILE, ABORT, filename);
    }
    n = fread (magicno, (size_t)1, (size_t)10, fp);
    fclose(fp);

    // If file too small
    if (n < 10) {
        return FMT_UNKNOWN;  
    }

    // Set file type
    type = FMT_UNKNOWN;   // default

    if ( magicno[0] == 'P' && magicno[1] >= '1' && magicno[1] <= '6' ) {
        type = FMT_PNM;
    } else if ( magicno[0] == 'W' && magicno[1] >= '1' && magicno[1] <= '4' ) {
        type = FMT_WNM;
    }
    return type;
}

//------------------------------------------------
int
ReadFile ( char *filename, const FileFormat filetype, Image *image )
{
    int rv = 0;
    switch (filetype) {
        case FMT_PNM:  rv = LoadPNM ( filename, image );  break;
        case FMT_WNM:  rv = LoadWNM ( filename, image );  break;
        default:       Error ( "ReadFile", NO_FORMAT, ABORT );
    }

    return rv;
}

//------------------------------------------------
int
WriteFile ( char *filename, Image *image )
{
    FILE *fp;
    byte *pic, *pix;
    int   i, j, w, h, rv;

    w = image->width;
    h = image->height;

    pic = NULL;
    if ( image->outFileType == F_PBMRAW) {
        if ( image->colorPlanes == 1 ) {
            pic = (byte *) calloc ((size_t)(w*h), (size_t)1);
            if ( !pic ) {
                Error ("WriteFile", CALLOC_FAIL, ABORT);
            }
            for ( i=0, pix=pic ; i<h ; i++ ) {
                for ( j=0 ; j<w ; j++, pix++ )
                    *pix = (byte)ROUND( image->band[GREY][i][j], 0, 255 );
            }
        } else if ( image->colorPlanes == 3 ) {
            pic = (byte *) calloc ((size_t)(w*h*3), (size_t)1);
            if ( !pic ) {
                Error ("WriteFile", CALLOC_FAIL, ABORT);
            }
            for ( i=0, pix=pic ; i<h ; i++ ) {
                for ( j=0 ; j<w ; j++, pix+=3 ) {
                    pix[0] = (byte)ROUND( image->band[RED][i][j], 0, 255 );
                    pix[1] = (byte)ROUND( image->band[GREEN][i][j], 0, 255 );
                    pix[2] = (byte)ROUND( image->band[BLUE][i][j], 0, 255 );
                }
            }
        }
    }

    // Open the file for writing 
    if ( (fp = fopen (filename, "wb")) == NULL ) {
        Error ("OpenFile", NO_FILE, ABORT, filename);
    }  
    
    rv = 0;
    switch (image->outFileType) {
        case F_PBMRAW:
            rv = WritePNM (fp, pic, w, h, image->colType, image->bpp,
                           image->comment);
            break;
        case F_WTFWLT:
            rv = WriteWNM (fp, image, w, h, image->colType,
	                   image->bpp, image->comment);
            break;
        default:
            Error ( "WriteFile", NO_FORMAT, ABORT );
    }

    fclose(fp);

    // Free memory 
    if (image->outFileType == F_PBMRAW) {
        free (pic);
    }
    return rv;
}

//------------------------------------------------
int
WritePNM ( FILE *fp, byte *pic, const int w, const int h, 
          const int colorstyle, const int bpp, char *comment )
{
    int   magic;
    byte *pix;
    int   i, j, len;
    char  buf[BUFSIZ];

    sprintf (buf , "%4dx%-4d %s, %s format.", w, h,
            (colorstyle==0) ? "PPM" : "PGM", "raw");
    fprintf (stdout, "\n[%s]", buf);

    // Calc magic number for this file 
    magic = 0;
    if (colorstyle==0) magic = 3;
    if (colorstyle==1) magic = 2;
    
    if (magic) magic += 3;

    // Write the header info
    fprintf (fp, "P%d\n",magic);
    fprintf (fp, "# Created By: %s\n", "MDR 1.0");
    if (comment) {  
        char *sp;
        sp = comment;
        while (*sp) {
            fprintf (fp, "# ");
            while (*sp && *sp != '\n') fputc((int)*sp++, fp);
            if (*sp == '\n') sp++;
            fputc('\n', fp);
        }
    }

    fprintf (fp, "%d %d\n", w, h);
    if ( colorstyle!=2 ) fprintf (fp, "255\n");

    if ( ferror(fp) ) return 0;

    // Write image data
    if ( colorstyle==0 ) {              // 24bit RGB, 3 bytes per pixel 
        for ( i=0, pix=pic, len=0; i<h; i++ ) {
            for ( j=0 ; j<w ; j++ ) {
                if ( bpp == 24 ) {
	                fputc((int)pix[0], fp);
                    fputc((int)pix[1], fp);
                    fputc((int)pix[2], fp);
                } else {
                    fputc((int)*pix, fp);
                    fputc((int)*pix, fp);
                    fputc((int)*pix, fp);
                }
                pix += (bpp==24) ? 3 : 1;
            }
        }
    } else if (colorstyle==1) {             // 8-bit greyscale 
        for ( i=0, pix=pic, len=0 ; i<w*h ; i++ ) {
            fputc ( (bpp==24) ? LUMA(pix[0],pix[1],pix[2]) : *pix, fp);
            pix += (bpp==24) ? 3 : 1;
        }
    }

    if ( ferror(fp) )
        return 0;

    return 1;
}

//------------------------------------------------
int
WriteWNM ( FILE *fp, Image *image, const int w, const int h, 
          const int colorstyle, const int bpp, 
          char *comment )
{
    int    magic;
    Vector pix, pixR, pixG, pixB;
    int    i;
    char   buf[BUFSIZ];

    sprintf (buf , "%4dx%-4d %s, %s format.", w, h,
            (colorstyle==0) ? "WPM" : "WGM", "wavelet");
    fprintf (stdout, "\n[%s]", buf);

    // Calc magic number for this file
    magic = 0;
    if (colorstyle==0) magic = 2;
    if (colorstyle==1) magic = 1;

    // Write the header info
    fprintf (fp, "W%d\n",magic);
    fprintf (fp, "# Created By: %s\n", "MDR 1.0");
    if (comment) {   
        char *sp;
        sp = comment;
        while (*sp) {
            fprintf (fp, "# ");
            while (*sp && *sp != '\n') fputc((int)*sp++, fp);
            if (*sp == '\n') sp++;
            fputc('\n', fp);
        }
    }

    fprintf (fp, "%d %d\n", w, h);

    if ( ferror(fp) )
        return 0;

    // Write image data 
    if ( colorstyle==0 ) {                  // 24bit RGB, 3 bytes per pixel
        pixR = image->band[RED][0];
        pixG = image->band[GREEN][0];
        pixB = image->band[BLUE][0];
        for ( i=0 ; i<w*h ; i++, pixR++, pixG++, pixB++ ) {
            int numwrote = 0;
            numwrote = fwrite( pixR, (size_t)sizeof(pixR[0]), (size_t)1, fp);
            numwrote += fwrite( pixG, (size_t)sizeof(pixG[0]), (size_t)1, fp);
            numwrote += fwrite( pixB, (size_t)sizeof(pixB[0]), (size_t)1, fp);
            if (numwrote < 3) {
                fclose(fp);
                Error("WriteWNM", WRITE_ERROR, RETURN);
                return 0;
            }
        }
    } else if (colorstyle==1) {             // 8-bit greyscale
        int height = h;
	    int width  = w;
        pix = image->band[GREY][0];
        while (height--) {
            if (fwrite( pix, (size_t)sizeof(pix[0]), (size_t)width, fp ) == 0) {
                fclose( fp );
                Error( "WriteWNM", WRITE_ERROR, RETURN );
                return 0;
            }
            pix += width;
        }
    }

    if ( ferror(fp) )
        return 0;

    return 1;
}

//------------------------------------------------
int
LoadPNM ( const char *filename, Image *image )
{
    FILE  *fp;
    int    c, c1;
    int    maxv, rv;

    maxv = rv = 0;

    image->band[RED]   = (Matrix)NULL;
    image->band[GREEN] = (Matrix)NULL;
    image->band[BLUE]  = (Matrix)NULL;
    image->comment     = (char *)NULL;

    if ( (fp = fopen (filename, "rb")) == NULL ) {
        Error ("OpenFile", NO_FILE, ABORT, filename);
    }  

    // check file length
    fseek (fp, 0L, 2);
    filesize = ftell(fp);
    fseek (fp, 0L, 0);


    // Determine file format:  "P1" = ascii bitmap, "P2" = ascii greymap,
    //   "P3" = ascii pixmap, "P4" = raw bitmap, "P5" = raw greymap,
    //   "P6" = raw pixmap. We only use P5 and P6.
    c = getc(fp);  c1 = getc(fp);
    if (c!='P' || c1<'5' || c1>'6') {
        Error ("LoadPNM", CUSTOM, RETURN, "Invalid magic number.\n");
        return 0;
    }

    // Read in header information 
    image->width  = getint(fp, image);
    image->height = getint(fp, image);

    // read the max value
    maxv = getint(fp, image);
    if (maxv < 1) {
        fclose(fp);
        if (image->comment) {
            free(image->comment);
        }
        image->comment = (char *)NULL;
        Error ("LoadPNM", CUSTOM, RETURN, "Garbage characters in header.\n");
        return 0;
    }
    image->outFileType = F_PBMRAW;

    if (c1=='5') {
        rv = loadpgm(fp, image, maxv);
    } 
    if (c1=='6') {
        rv = loadppm(fp, image, maxv);
    }
    fclose(fp);

    if (!rv) {
        if (image->band[RED]) {
            FreeArrayMemory(image->band[RED]);
        }
        if (image->band[GREEN]) {
            FreeArrayMemory(image->band[GREEN]);
        }
        if (image->band[BLUE]) {
            FreeArrayMemory(image->band[BLUE]);
        }
        if (image->comment) {
            free(image->comment);
        }
        image->band[RED]   = (Matrix)NULL;
        image->band[GREEN] = (Matrix)NULL;
        image->band[BLUE]  = (Matrix)NULL;
        image->comment     = (char *)NULL;
    }

    return rv;
}  


//------------------------------------------------
static int loadpgm ( FILE *fp, Image *image, int maxv )
{
    Vector  pixFlt;
    byte   *pic8;
    int     i, j, bitshift, w, h, holdmaxv, rv;

    // Init image
    rv = (int)AllocImageMemory (image, (long)image->width, (long)image->height, 1, 8);
    if ( !rv ) {
        Error ("loadpgm", MEMORY_IMAGE, ABORT);
    }
    image->colType = 1;   // Greyscale
    w = image->width;
    h = image->height;
    sprintf(image->fullInfo, "%dx%d PGM, %s format.(%ld bytes)", 
	    width, height, "raw", filesize);

    // if maxv > 255, drop bits
    holdmaxv = maxv;
    bitshift = 0;
    while ( maxv>255 ) { maxv = maxv>>1;  bitshift++; }

    numgot = 0;

    if ( holdmaxv>255 ) {
        for ( i=0, pixFlt=image->band[GREY][0] ; i<h ; i++ ) {
            for ( j=0 ; j<w ; j++, pixFlt++ ) {
                *pixFlt = (float) (getshort(fp) >> bitshift);
            }
        }
    } else {    // read raw
        pic8 = (byte *) calloc((size_t)(w*h), (size_t)1);
        if ( !pic8 ) {
            Error ("loadpgm", CALLOC_FAIL, ABORT);
        }
        numgot = (long)fread(pic8, (size_t)1, (size_t)(w*h), fp);
        for ( i=0, pixFlt=image->band[GREY][0] ; i<w*h ; i++, pixFlt++ ) {
            *pixFlt = (float)pic8[i];
        }
        free (pic8);
    }

    if ( numgot != (long)(w*h) ) {
        Error ("loadpgm", CUSTOM, RETURN, TRUNCSTR);  
    }

    if (garbage) {
        Error ("loadpgm", CUSTOM, RETURN, "Garbage characters in image data.");
        return 0;
    }

    return 1;
}

//------------------------------------------------
static int loadppm ( FILE *fp, Image *image, int maxv )
{
    byte *pix, *pic24;
    byte scale[256];
    int   i, j, bitshift, w, h, holdmaxv, rv;

    w = image->width;
    h = image->height;

    // allocate 24-bit image
    pic24 = (byte *) calloc((size_t)(w*h*3), (size_t) 1);
    if ( !pic24 ) {
        Error ("loadppm", CALLOC_FAIL, ABORT);
    }

    // init image
    rv = (int)AllocImageMemory (image, (long)image->width, (long)image->height, 3, 24);
    if ( !rv ) {
        Error ("loadppm", MEMORY_IMAGE, ABORT);
    }
    image->colType = 0;     // color
    sprintf(image->fullInfo, "PPM, %s format.  (%ld bytes)", 
	    "raw", filesize);
    sprintf(image->shortInfo, "%dx%d PPM.", w, h);

    // if maxv > 255, drop bits
    holdmaxv = maxv;
    bitshift = 0;
    while ( maxv>255 ) { maxv = maxv>>1;  bitshift++; }

    numgot = 0;

    if ( holdmaxv>255 ) {
        for ( i=0, pix=pic24 ; i<h ; i++ ) {
            for ( j=0; j<w*3; j++, pix++ )
                *pix = (byte) (getshort(fp) >> bitshift);
        }
    } else {   // read data
        numgot = (long)fread(pic24, (size_t)1, (size_t)(w*h*3), fp);
    }
  
    if ( numgot != (long)(w*h*3) ) {
        Error ("loadppm", CUSTOM, RETURN, TRUNCSTR);
    }

    if (garbage) {
        Error ("loadppm", CUSTOM, RETURN, "Garbage characters in image data.");
        return 0;
    }

    // Scale all RGB values to 0-255 
    if (maxv < 255) { 
        for ( i=0 ; i<=maxv ; i++)
            scale[i] = (byte)( (i * 255) / maxv );

        for ( i=0, pix=pic24 ; i<h ; i++ ) {
            for ( j=0 ; j<w*3 ; j++, pix++ )
                *pix = scale[*pix];
        }
    }

    // Copy byte values to image bands 
    for ( i=0, pix=pic24 ; i<h ; i++ ) {
        for ( j=0 ; j<w ; j++, pix+=3 ) {
            image->band[RED][i][j]   = (float)(pix[0]);
            image->band[GREEN][i][j] = (float)(pix[1]);
            image->band[BLUE][i][j]  = (float)(pix[2]);
        }
    }
    free ( (byte *)pic24 );

    return 1;
}

//------------------------------------------------
int
LoadWNM ( const char *filename, Image *image )
{
    FILE  *fp;
    int    c, c1;
    int    rv;

    /* Set initial values */
    garbage = rv = 0;
    image->band[RED]   = (Matrix)NULL;
    image->band[GREEN] = (Matrix)NULL;
    image->band[BLUE]  = (Matrix)NULL;
    image->comment     = (char *)NULL;

    if ( (fp = fopen (filename, "rb")) == NULL ) {
        Error ("OpenFile", NO_FILE, ABORT, filename);
    }  

    // check file length
    fseek (fp, 0L, 2);
    filesize = ftell(fp);
    fseek (fp, 0L, 0);

    // file format : "W1" = wavelet greymap,"W2" = wavelet pixmap 
    c = getc(fp);  c1 = getc(fp);
    if (c!='W' || c1<'1' || c1>'2') {
        Error ("LoadWNM", CUSTOM, RETURN, "Invalid magic number.\n");
        return 0;
    }

    // read header
    image->width  = getint(fp, image);
    image->height = getint(fp, image);
    image->outFileType = F_WTFWLT;

    if ( c1=='1')
        rv = loadwgm( fp, image);
    else if ( c1=='2')
        rv = loadwpm( fp, image);

    fclose(fp);

    if (!rv) {
        if (image->band[RED]) {
            FreeArrayMemory(image->band[RED]);
        }
        if (image->band[GREEN]) {
            FreeArrayMemory(image->band[GREEN]);
        }
        if (image->band[BLUE]) {
            FreeArrayMemory(image->band[BLUE]);
        }
        if (image->comment) {
            free(image->comment);
        }
        image->band[RED]   = (Matrix)NULL;
        image->band[GREEN] = (Matrix)NULL;
        image->band[BLUE]  = (Matrix)NULL;
        image->comment     = (char *)NULL;
    }
    return rv;
}

//------------------------------------------------
static int loadwgm ( FILE *fp, Image *image)
{
    Vector  pixFlt;
    int     w, h, rv;

    // init image
    rv = (int)AllocImageMemory (image, (long)image->width, (long)image->height, 1, 8);
    if ( !rv ) {
        Error ("loadwgm", MEMORY_IMAGE, ABORT);
    }
    image->colType = 1;   // grey scale
    w = image->width;
    h = image->height;
    sprintf(image->fullInfo, "WGM, %s format.  (%ld bytes)", 
            "wavelet", filesize);
    sprintf(image->shortInfo, "%dx%d WGM.", image->width, image->height);

    numgot = 0;
    pixFlt = image->band[GREY][0];

    numgot = (long)fread(pixFlt, (size_t)sizeof(pixFlt[0]), (size_t)(w*h), fp);
    if ( numgot != (long)(w*h) ) {
        Error ("loadwgm", CUSTOM, RETURN, TRUNCSTR);  
    }

    if (garbage) {
        Error ("loadwgm", CUSTOM, RETURN, "Garbage characters in image data.");
        return 0;
    }
    return 1;
}

//------------------------------------------------
static int loadwpm ( FILE *fp, Image *image)
{
    Vector pixR, pixG, pixB;
    int   i, w, h, rv;

    w = image->width;
    h = image->height;

    // init image
    rv = (int)AllocImageMemory (image, (long)image->width, (long)image->height, 3, 24);
    if ( !rv ) {
        Error ("loadwpm", MEMORY_IMAGE, ABORT);
    }
    image->colType = 0;   // color
    sprintf(image->fullInfo, "WPM, %s format.  (%ld bytes)", 
	    "wavelet", filesize);
    sprintf(image->shortInfo, "%dx%d WPM.", w, h);

    numgot = 0;

    pixR = image->band[RED][0];
    pixG = image->band[GREEN][0];
    pixB = image->band[BLUE][0];
    for ( i=0 ; i<w*h*3 ; i+=3, pixR++, pixG++, pixB++ ) {
        numgot += (long)fread( pixR, (size_t)sizeof(pixR[0]), (size_t)1, fp );
        numgot += (long)fread( pixG, (size_t)sizeof(pixG[0]), (size_t)1, fp );
        numgot += (long)fread( pixB, (size_t)sizeof(pixB[0]), (size_t)1, fp );
    }
  
    if ( numgot != (long)(w*h*3) ) {
        Error ("loadwpm", CUSTOM, RETURN, TRUNCSTR);
    }

    if (garbage) {
        Error ("loadwpm", CUSTOM, RETURN, "Garbage characters in image data.");
        return 0;
    }
    return 1;
}

//------------------------------------------------
static int getint ( FILE *fp, Image *image )
{
    int c, i, firstchar;

    // skip to start of next number
    c = getc(fp);
    while (1) {
        // ignore comments
        if ( c=='#' ) {   
            char cmt[256], *sp, *tmpptr;
            sp = cmt; firstchar = 1;
            while (1) {
    	        c = getc(fp);
                if ( firstchar && c == ' ' ) {
                    firstchar = 0;  
                } else {
	                if ( c == '\n' || c == EOF ) break;
	                if ( (sp-cmt)<250 ) *sp++ = (char)c;
	            }
            }
            *sp++ = '\n';
            *sp   = '\0';

            if ( strlen(cmt) > 0 ) {    // fill image comment
	            if ( !image->comment ) {
	                image->comment = (char *) malloc(strlen(cmt) + 1);
                    if ( !image->comment ) {
                        Error ("getint", MALLOC_FAIL, ABORT);
                    }
	                image->comment[0] = '\0';
	            } else {
	                tmpptr = (char *) realloc(image->comment, 
		            strlen(image->comment) + strlen(cmt) + 1); 
                    if (!tmpptr) {
                        Error ("getint", REALLOC_FAIL, ABORT);
                    }
	                image->comment = tmpptr;
	            }
	            strcat(image->comment, cmt);
            }
        }

        if ( c==EOF )
            return 0;
        if ( c>='0' && c<='9' )
            break;   

        // any garbage? 
        if ( c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',') garbage=1;

        c = getc(fp);
    }

    i = 0;
    while (1) {
        i = (i*10) + (c - '0');
        c = getc(fp);
        if ( c==EOF )
            return i;
        if ( c<'0' || c>'9' )
            break;
    }

    numgot++;
    return i;
}

//------------------------------------------------
static int getshort ( FILE *fp )
{
    int c1, c2;

    c1 = getc(fp);
    if ( c1 == EOF )
        return 0;
    c2 = getc(fp);
    if ( c2 == EOF )
        return 0;

    numgot++;

    return (c2 << 8) | c1;
}

//------------------------------------------------
static int getbit ( FILE *fp, Image *image )
{
    int c;

    // skip to start of next number
    c = getc(fp);
    while (1) {
        if ( c=='#' ) {   
            char cmt[256], *sp, *tmpptr;

            sp = cmt;
            while (1) {
	        c = getc(fp);
	        if ( c == '\n' || c == EOF )
                    break;

	        if ( (sp-cmt)<250 )
                    *sp++ = (char)c;
            }
            *sp++ = '\n';
            *sp   = '\0';

            if ( strlen(cmt) > 0 ) {    
	            if ( !image->comment ) {
	                image->comment = (char *) malloc(strlen(cmt) + 1);
                    if ( !image->comment ) {
                        Error ("getbit", MALLOC_FAIL, ABORT);
                    }
	                image->comment[0] = '\0';
                } else {
	                tmpptr = (char *) realloc(image->comment, 
		            strlen(image->comment) + strlen(cmt) + 1); 
                    if (!tmpptr) {
                        Error ("getbit", REALLOC_FAIL, ABORT);
                    }
	                image->comment = tmpptr;
	            }
	            strcat(image->comment, cmt);
            }
        }
        if ( c==EOF )
            return 0;
        if ( c=='0' || c=='1' )
            break;  

        if ( c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',' )
            garbage=1;

        c = getc(fp);
    }

    numgot++;
    return (c-'0');
}

