
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <errors.h>   
#include <global.h>   
#include <lifting.h>      
#include <image.h>       
#include <util.h>    

#define EPSILON 1.0e-20;  // number close to zero (for LU)

static Matrix momentX = NULL;   /* Integral and Moments in X direction */
static Matrix momentY = NULL;   /* Integral and Moments in Y direction */

static void UpdateMomentsMatrix ( Matrix, const Vector, const int, const int,
                                   const int, const int, const int );

static void UpdateLifting ( Vector, const Matrix, const int, const int,
                                    const int, const int );

//------------------------------------------------
bool
InitMomentsMatrix ( const int R, const int sizeX, const int sizeY,
             const bool print )
{
    int row, col;

    if ( ODD(R) || R<0 )
        return FALSE;

    momentX = AllocFloatArray( 0, (long)(sizeX-1), 0, (long)(R-1) );
    momentY = AllocFloatArray( 0, (long)(sizeY-1), 0, (long)(R-1) );

    // Initialize integral and moments
    for ( row=0 ; row<sizeY ; row++ ) {
        for ( col=0 ; col<R ; col++ ) {
            if ( row==0 && col==0 ) {
                momentY[row][col] = (float)1.0;   
            } else {
                momentY[row][col] = (float)pow( (float)row, (float)col );
            }
        }
    }
    for ( row=0 ; row<sizeX ; row++ ) {
        for ( col=0 ; col<R ; col++ ) {
            if ( row==0 && col==0 ) {
                momentX[row][col] = (float)1.0;   
            } else {
                momentX[row][col] = (float)pow( (float)row, (float)col );
            }
        }
    }
    // Print moment coefficients
    if (print) {
        fprintf (stdout, "Integral-Moments Coefficients for X (R=%2d)\n", R);
        for ( row=0 ; row<sizeX ; row++ ) {
            fprintf (stdout, "Coeff[%d] (%.20g", row, momentX[row][0]);
            for ( col=1 ; col<R ; col++ ) {
                fprintf (stdout, ",%.20g", momentX[row][col]);
            }
            fprintf (stdout, ")\n");
        }
        fprintf (stdout, "Integral-Moments Coefficients for Y (R=%2d)\n", R);
        for ( row=0 ; row<sizeY ; row++ ) {
            fprintf (stdout, "Coeff[%d] (%.20g", row, momentY[row][0]);
            for ( col=1 ; col<R ; col++ ) {
                fprintf (stdout, ",%.20g", momentY[row][col]);
            }
            fprintf (stdout, ")\n");
        }
    }

    return TRUE;
}

//------------------------------------------------
void
FreeMomentsMatrix ( const int sizeX, const int sizeY, const int R )
{
    FreeFloatArray( momentX, 0, (long)(sizeX-1), 0, (long)(R-1) );
    FreeFloatArray( momentY, 0, (long)(sizeY-1), 0, (long)(R-1) );
}

//------------------------------------------------
void
GetLifting ( Vector lifting, const Vector filter, const int step,
                             const int N, const int R, const int L,
                             const bool direction, const bool type )
{
    Matrix moment;   
    int len, step2, noGammas;    

    if (!lifting) {
        Error ("GetLifting", NO_LIFTING_MATRIX, ABORT);
    }
    // select proper moments matrix
    if (!direction) {   
        moment = momentX;
    } else {
        moment = momentY;
    }

    len = CEIL (L, step);
    step2 = step << 1;
    noGammas  = len >> 1; 

    if ( type == false ) { 
        UpdateMomentsMatrix ( moment, filter, step2, noGammas, len, N, R );
        UpdateLifting ( lifting, moment, len, step2, noGammas, R );
    } else { 
        UpdateLifting ( lifting, moment, len, step2, noGammas, R );
        UpdateMomentsMatrix ( moment, filter, step2, noGammas, len, N, R );
    }
}

//------------------------------------------------
static void
UpdateMomentsMatrix ( Matrix moment, const Vector filter, const int step2,
                      const int noGammas, const int len,
                      const int N, const int R )
{
    int i, j, row, col, idxL, idxG;
    int LlessthanR, LequalR, LgreaterthanR; 
    Vector filterPtr;      

    LlessthanR = (N>>1) - 1;               
    LgreaterthanR = (N>>1) - ODD(len);          
    LequalR = noGammas - (LlessthanR + LgreaterthanR);  

    // left lambdas less than right
    filterPtr = filter;
    idxG = step2 >> 1;
    for ( row = 1 ; row <= LlessthanR ; row++ ) {
        idxL = 0;
        filterPtr += N;
        for ( col = 0 ; col < N ; col++ ) {
            for ( j = 0 ; j < R ; j++ ) {
                moment[idxL][j] += filterPtr[col]*moment[idxG][j];
            }
            idxL += step2;
        }
        idxG += step2;
    }
    // left lambdas equal right
    filterPtr += N;
    for ( i = 0 ; i < LequalR ; i++ ) {
        idxL = i*step2;
        for ( col = 0 ; col < N ; col++ ) {
            for ( j = 0 ; j < R ; j++ ) {
                moment[idxL][j] += filterPtr[col]*moment[idxG][j];
            }
            idxL += step2;
        }
        idxG += step2;   
                        
    }
    // left lambdas > right
    for ( row = LgreaterthanR ; row >= 1 ; row-- ) {
        idxL = (LequalR-1)*step2;   
        filterPtr -= N;  
        for ( col = N-1 ; col >= 0 ; col-- ) {
            for ( j = 0 ; j < R ; j++ ) {
                moment[idxL][j] += filterPtr[col]*moment[idxG][j];
            }
            idxL += step2;
        }
        idxG += step2;  
    }
}

//------------------------------------------------
static void
UpdateLifting ( Vector lifting, const Matrix moment, const int len,
                                const int step2, const int noGammas, const int R)
{
    int lcIdx, i, j, row, col,  idxL, idxG;
    int LlessthanR, LequalR, LgreaterthanR; 
    Matrix lift; 
    Vector b; 
    int *indx;  
    float d;  

    // calc lifting coefficients
    lcIdx = 0;    
    indx = AllocIntVector( 1, (long)R );
    lift = AllocFloatArray( 1, (long)R, 1, (long)R );  
    b = AllocFloatVector( 1, (long)R );   

    LlessthanR = (R>>1) - 1; 
    LgreaterthanR = (R>>1) - ODD(len);       
    LequalR = noGammas - (LlessthanR + LgreaterthanR);

    // left lambdas less than right
    idxG = step2>>1;  
    for ( i=0 ; i<LlessthanR ; i++ ) {
        idxL = 0;  
        for ( col=1 ; col<=R ; col++ ) {
            for ( row=1 ; row<=R ; row++ )
                lift[row][col] = moment[idxL][row-1];    
            idxL += step2;
        }
        // apply LU decomposition to lift coeffs
        LUDecompose (lift, R, indx, &d);
        for ( j=1 ; j<=R ; j++) {
            b[j] = moment[idxG][j-1]; 
        }
        // do back substitution to get lifting coeffs
        LUBacksubstitute (lift, R, indx, b);
        for ( col=1 ; col<=R ; col++ ) { 
            lifting[lcIdx++] = b[col];
        }
        idxG += step2; 
    }

    // left lambdas equal right
    for ( i=0 ; i<LequalR ; i++ ) {
        idxL = i*step2;
        for ( col=1 ; col<=R ; col++ ) {
            for ( row=1 ; row<=R ; row++ )
                lift[row][col] = moment[idxL][row-1];     
            idxL += step2;
        }
        LUDecompose (lift, R, indx, &d);
        for ( j=1 ; j<=R ; j++) {
            b[j] = moment[idxG][j-1]; 
        }
        LUBacksubstitute (lift, R, indx, b);
        for ( col=1 ; col<=R ; col++ ) {
            lifting[lcIdx++] = b[col];
        }
        idxG += step2;  
    }

    // left lambdas greater than right
    for ( i=0 ; i<LgreaterthanR ; i++ ) {
        idxL = (LequalR-1)*step2;  
        for ( col=1 ; col<=R ; col++ ) {
            for ( row=1 ; row<=R ; row++ )
                lift[row][col] = moment[idxL][row-1];     
            idxL += step2;
        }
        LUDecompose (lift, R, indx, &d);
        for ( j=1 ; j<=R ; j++) {
            b[j] = moment[idxG][j-1];   
        }
        LUBacksubstitute (lift, R, indx, b);
        for ( col=1 ; col<=R ; col++ ) {  
            lifting[lcIdx++] = b[col];
        }
        idxG += step2; 
    }
    FreeFloatArray( lift, 1, (long)R, 1, (long)R );
    FreeFloatVector( b, 1, (long)R );
    FreeIntVector( indx, 1, (long)R );
}

//------------------------------------------------
float
Neville ( const float *x, const float *f, const int n, const float xx )
{
    register int i,j;
    Vector vy;
    float y;

    vy = AllocFloatVector( 0, (long)(n-1) );   
    for ( i=0; i<n; i++ ) {
        vy[i] = f[i];
        for ( j=i-1; j>=0; j--) {
            float den = x[i] - x[j];
            if ( zero(den) ) {
                Error ("Neville", DENOM_NEAR_ZERO, RETURN);
            }
            vy[j] = vy[j+1] + (vy[j+1] - vy[j]) * (xx - x[i]) / den;
        }
    }
    y = vy[0];
    FreeFloatVector( vy, 0, (long)(n-1) );

    return y;
}

//------------------------------------------------
Vector
GetFilter ( const int N )
{
    static float xa[40], ya[40], x;
    static int row, col, cc, Nrows;
    Vector filter, ptr;

    assert ( N < sizeof xa / sizeof *xa ); 
    if ( ODD (N) || (N == 0) ) {
        return NULL;   
    }
    Nrows = (N>>1) + 1;    

    filter = AllocFloatVector( 0, (long)(Nrows*N - 1) );
    ptr = filter;

    xa[0] = (float)0.5*(float)(1 - N);   
    for (col = 1 ; col < N ; col++)
        xa[col] = xa[col-1] + (float)1;

    filter += ( (Nrows*N) - 1 ); 
    for (row = 0 ; row < Nrows ; row++) {
        x = (float)row;
        for (col = 0 ; col < N ; col++) {
            for (cc = 0 ; cc < N ; cc++)
                ya[cc] = (float)0;
            ya[col] = (float)1;
            *(filter--) = Neville (xa, ya, N, x);
        }
    }
    return ptr;
}

//------------------------------------------------
void 
LUDecompose ( Matrix a, int n, int *indx, float *d )
{
    int i, imax, j, k;
    float big, dum, sum, temp;
    Vector vv;  

    vv=AllocFloatVector( 1, (long)n );
    *d=(float)1;              
    for (i=1;i<=n;i++) {   
        big=(float)0;       
        for (j=1;j<=n;j++)
            if ((temp=(float)fabs((float)a[i][j])) > big)
                big=temp;
        if (big == (float)0.0)
            Error ("LUdmcp", SINGULAR_MATRIX, ABORT);
        vv[i]=(float)1/big;   
    }
    for (j=1;j<=n;j++) {   // Crout's loop
        for (i=1;i<j;i++) {  
            sum=a[i][j];     
            for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
            a[i][j]=sum;
        }
        big=(float)0;   // search for pivot element
        imax = 0;     
        for (i=j;i<=n;i++) {  
            sum=a[i][j];      
            for (k=1;k<j;k++)
                sum -= a[i][k]*a[k][j];
            a[i][j]=sum;
            if ( (dum=vv[i]*(float)fabs((float)sum)) >= big) {
                big=dum;
                imax=i;
            }
        }
        if (j != imax) {       // interchange rows?
            for (k=1;k<=n;k++) { 
                dum=a[imax][k];
                a[imax][k]=a[j][k];
                a[j][k]=dum;
            }
            *d = -(*d);         
            vv[imax]=vv[j];      
        }
        indx[j]=imax;
        if (a[j][j] == (float)0.0)
            a[j][j]=(float)EPSILON;
        if (j != n) {       // divide by pivot
            dum=(float)1/(a[j][j]);
            for ( i=j+1 ; i<=n ; i++ )
                a[i][j] *= dum;
        }
    }  
    FreeFloatVector( vv, 1, (long)n);
}
    
#undef EPSILON

//------------------------------------------------
void 
LUBacksubstitute ( Matrix a, int n, int *indx, Vector b )
{
    int i,ii=0,ip,j;
    float sum;

    for (i=1;i<=n;i++) {  
        ip=indx[i];       
        sum=b[ip];         
        b[ip]=b[i];        
        if (ii)            
            for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
        else if (sum)     
            ii=i;         
        b[i]=sum;
    }
    for (i=n;i>=1;i--) {  
        sum=b[i];
        for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
        b[i]=sum/a[i][i];  
        if (zero(b[i])) b[i] = (float)0;  
    }                     
}

