
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <constant.h>    
#include <file.h>        
#include <transform.h>     
#include <util.h>     
#include <errors.h>   
#include <lifting.h>
#include <image.h>
#include <global.h>

static Vector filter = NULL;      // Vector with filter coefficients 
static Matrix liftingX = NULL;    // Matrix with lifting coefficients in X 
static Matrix liftingY = NULL;    // Matrix with lifting coefficients in Y

static void PrintMatrix ( const Matrix, const int, const int );
static void Point ( Image *, const int, const int, const int, const int );
static void Scale ( Matrix, const int, const int, const int,
                            const int, const int );
static void MinMax ( const Matrix, const int, const int, float *, float * );

//------------------------------------------------
void
LiftingWaveletTranform (ParamStruct param)
{
    int        i;
    Image      image;          
    FileFormat filetype;       
    float    startclock, stopclock;    
    float        diff, total=(float)0;

    // Read image from input file
 
    // Read file type 
    fprintf (stdout, "Read file type ");
    fflush (stdout);
    
    startclock = get_current_time();
    filetype = ReadFileType (param.infile);
    stopclock = get_current_time();
    diff = ((float)stopclock-(float)startclock);

    if ( filetype == FMT_ERROR ) {
        Error ("LiftingWaveletTranform", NO_FILE, ABORT, param.infile);
    } else if ( filetype == FMT_UNKNOWN ) {
        Error ("LiftingWaveletTranform", INVALID_FILE_FORMAT, ABORT, param.infile);
    } else if ( filetype < FMT_ERROR ) {
        Error ("LiftingWaveletTranform", NO_READABLE_FORMAT, ABORT, param.infile);
    }
    fprintf (stdout, "\t\t\t\t[%3.2f secs]\n", diff);
    total += diff;

    // Read input file of image
    fprintf (stdout, "Read input file");
    fflush (stdout);

    startclock = get_current_time();
    i = ReadFile (param.infile, filetype, &image);
    stopclock = get_current_time();
    diff = ((float)stopclock-(float)startclock);

    if ( !i ) {
        Error ("LiftingWaveletTranform", NO_FILE, ABORT, param.infile);
    }
    param.sizeX = image.width;
    param.sizeY = image.height;
    if ( image.width == 0 || image.height == 0 ) {  
        Error ("LiftingWaveletTranform", INVALID_RELATIONAL_INT, ABORT, 
               "Dimensions", ">", 0);
    }
    fprintf (stdout, "\t[%4dx%-4d]\t\t[%3.2f secs]\n",
             image.width, image.height, diff);
    total += diff;
    fprintf (stdout, "[%s]\n", image.fullInfo);

    SetOutputType( param.outFileType, filetype, &image);

    // print original data   
    if (param.print) {
        fprintf (stdout, "Original Data Set\n");
        for ( i=0 ; i<image.colorPlanes ; i++ ) {
            fprintf (stdout, "Band #%d\n", i);
            PrintMatrix (image.band[i], image.width, image.height);
        }
    }

    // find filter coefficients
    if (EVEN(param.N) ) {
	    fprintf (stdout, "\nFilter Coefficients\t\t\t");
	    fflush (stdout);
	    startclock = get_current_time();
	    GetDualWavelet (image.width, image.height, param.N, param.nTilde,
		                param.print, param.levels);
	    stopclock = get_current_time();
	    diff = ((float)stopclock-(float)startclock);
	    fprintf (stdout, "[%3.2f secs]\n", diff);
	    total += diff;
    }

    // find lifting coefficients 
    if (EVEN(param.N) ) {
        fprintf (stdout, "Lifting Coefficients\t\t\t");
        fflush (stdout);
        startclock = get_current_time();
        GetMoments (image.width, image.height, param.N, param.nTilde,
                     param.print, param.levels);
        stopclock = get_current_time();
        diff = ((float)stopclock-(float)startclock);
        fprintf (stdout, ">> %3.2f secs\n", diff);
        total += diff;
    }

    // Forward transform
    if (param.mode == FORWARD) {
        fprintf (stdout, "\nForward Wavelet Transform\t\t");
        fflush (stdout);
        startclock = get_current_time();
        for (i=0 ; i<image.colorPlanes ; i++) {
	        if (ODD(param.N) ) {
	            HaarLifting2DWT( image.band[i], image.width, image.height,
                            param.levels, false );
            } else {
                Lifting2DWT( image.band[i], image.width, image.height,
                            param.N, param.nTilde, param.levels, false );
            }
        }
        stopclock = get_current_time();
        diff = ((float)stopclock-(float)startclock);
        fprintf (stdout, "[%3.2f secs]\n", diff);
        total += diff;
    }

    // Print wavelet coefficients
    if (param.print && param.mode != INVERSE) {
        fprintf (stdout, "Wavelet Coefficients\n");
        for ( i=0 ; i<image.colorPlanes ; i++ ) {
            fprintf (stdout, "Band #%d\n", i);
            PrintMatrix (image.band[i], image.width, image.height);
        }
    }

    // reorganize into Mallat form
    if (param.mode == FORWARD && param.mallat) {
        fprintf (stdout, "Organize in Mallat format\t\t");
        fflush (stdout);
        startclock = get_current_time();
            for ( i=0 ; i<image.colorPlanes ; i++ ) {
                LiftingToMallat (image.band[i], image.width, image.height,
                                  param.N, param.nTilde, param.levels);
            }
        stopclock = get_current_time();
        diff = ((float)stopclock-(float)startclock);
        fprintf (stdout, "[%3.2f secs]\n", diff);
        total += diff;
    }

    // Inverse transform
    if (param.mode != FORWARD) {
        if (param.mallat) {
            fprintf (stdout, "Organize in Lifting format\t\t");
            fflush (stdout);
            startclock = get_current_time();
            for ( i=0 ; i<image.colorPlanes ; i++ ) {
                 MallatToLifting (image.band[i], image.width, image.height,
                                  param.N, param.nTilde, param.levels);
            }
            stopclock = get_current_time();
            diff = ((float)stopclock-(float)startclock);
            fprintf (stdout, "[%3.2f secs]\n", diff);
            total += diff;
        }
        fprintf (stdout, "\nInverse Wavelet Transform\t\t");
        fflush (stdout);
        startclock = get_current_time();
        for (i=0 ; i<image.colorPlanes ; i++ ) {
		    if (ODD(param.N)) {
                HaarLifting2DWT ( image.band[i], image.width, image.height,
                             param.levels, TRUE );
		    } else {
                Lifting2DWT (image.band[i], image.width, image.height,
                        param.N, param.nTilde, param.levels, TRUE);
		    }
        }
        stopclock = get_current_time();
        diff = ((float)stopclock-(float)startclock);
        fprintf (stdout, "[%3.2f secs]\n", diff);
        total += diff;
    }

    // scale results
    if (param.scale) {
        fprintf (stdout, "Scale image ");
        fflush (stdout);
        startclock = get_current_time();
        for ( i=0 ; i<image.colorPlanes ; i++ ) {
             Scale (image.band[i], image.width, image.height,
                    0, 255, param.gamma);
        }
        stopclock = get_current_time();
        diff = ((float)stopclock-(float)startclock);
        fprintf (stdout, "[%3.2f secs]\n", diff);
        fflush (stdout);
        total += diff;
    }

    // Print results
    if (param.print) {
        fprintf (stdout, "Final Coefficients\n");
        for (i=0 ; i<image.colorPlanes ; i++ ) {
            fprintf (stdout, "Band #%d\n", i);
            PrintMatrix (image.band[i], image.width, image.height);
        }
    }

    // write results
    fprintf (stdout, "\nWrite output file");
    fflush (stdout);
    startclock = get_current_time();
    i = WriteFile (param.outfile, &image);
    stopclock = get_current_time();
    if ( !i ) {
        Error ("LiftingWaveletTranform", WRITE_ERROR, ABORT);
    }
    diff = ((float)stopclock-(float)startclock);
    fprintf (stdout, "\t\t>> %3.2f secs\n", diff);
    total += diff;

    fprintf (stdout, "\t\t\t\t       -------------\n");
    fprintf (stdout, "Total execution time\t\t\t>> %3.2f secs\n", total);
    fflush (stdout);

    if ( EVEN(param.N) ) {
        FreeMemory (param.sizeX, param.sizeY, param.N, param.nTilde, 
                    param.levels);
    }
    FreeImageMemory (image);
}

//------------------------------------------------
void
GetDualWavelet ( const int width, const int height, const int N, const int nTilde,
          const bool print, const int levels )
{
    int row, col, noGammas, nX, nY;
    float maxN;
    Vector filterPtr;

    if ( (filter = GetFilter(N)) == NULL ) {
        Error ("GetDualWavelet", NOT_EVEN, RETURN, "N");
        Error ("GetDualWavelet", INVALID_RELATIONAL_INT, ABORT, "N", ">", 0);
    } else {
        if (print) {
            int Nrows = (N>>1) + 1;
            filterPtr = filter;
            fprintf (stdout, "\nFilter Coefficients (N=%d)\n", N);
            for (row = 0 ; row < Nrows ; row++) {
                fprintf (stdout, "For %d in the left and %d in the right\n[ ",
                         row, N-row);
                for (col = 0 ; col < N ; col++)
                    fprintf (stdout, "%.18g ", *(filterPtr++));
                fprintf (stdout, "]\n");
            }
        }
    }

    // Allocate lifting coefficient matrices 
    maxN = (float)MAX(N, nTilde) - (float)1;    // max vanishing moments 
    noGammas = width >> 1;                  // number of Xdim Gammas
    if (noGammas > 0) {
        nX = (int)logBaseN ( (float)(width-1)/maxN, (float)2 ); // x iters
        nX = (nX < 0) ? 0 : MIN(levels, nX);                // lowest level
        liftingX = AllocFloatArray( 0, (long)(nX-1), 0, (long)(noGammas*nTilde - 1) );
        for ( row=0 ; row<nX ; row++ )
            for ( col=0 ; col<noGammas*nTilde ; col++ )
                liftingX[row][col] = (float)0;
    }
    noGammas = height >> 1;    // Ydim Gammas
    if (noGammas > 0) {
        nY = (int)logBaseN ( (float)(height-1)/maxN, (float)2 );   
        nY = (nY < 0) ? 0 : MIN(levels, nY);   
        liftingY = AllocFloatArray( 0, (long)(nY-1), 0, (long)(noGammas*nTilde - 1) );
        for ( row=0 ; row<nY ; row++ )
            for ( col=0 ; col<noGammas*nTilde ; col++ )
                liftingY[row][col] = (float)0;
    }

    // integral moment matrices
    if (!InitMomentsMatrix (nTilde, width, height, print)) {
        Error ("GetDualWavelet", NOT_EVEN, RETURN, "N");
        Error ("GetDualWavelet", INVALID_RELATIONAL_INT, ABORT, "N", ">", 0);
    }
}

//------------------------------------------------
void
FreeMemory ( const int width, const int height, const int N, const int nTilde,
            const int levels )
{
    int noGammas, nRows, nX, nY;
    float maxN;

    // free filter coefficients
    nRows = (N>>1) + 1;
    FreeFloatVector( filter, 0, (long)(nRows*N - 1) );

    // free lifting coefficients
    maxN = (float)MAX(N, nTilde) - (float)1;   // max vanishing moments 
    if (liftingX) {
        noGammas = width >> 1;
        nX = (int)logBaseN( (float)(width-1)/maxN, (float)2 ); 
        nX = (nX < 0) ? 0 : MIN(levels, nX);   
        FreeFloatArray( liftingX, 0, (long)(nX-1), 0, (long)(noGammas*nTilde - 1) );
    }
    if (liftingY) {
        noGammas = height >> 1;
        nY = (int)logBaseN ( (float)(height-1)/maxN, (float)2 );  
        nY = (nY < 0) ? 0 : MIN(levels, nY);  
        FreeFloatArray( liftingY, 0, (long)(nY-1), 0, (long)(noGammas*nTilde - 1) );
    }

    // free integral moments 
    FreeMomentsMatrix( width, height, nTilde );
}

//------------------------------------------------
void
GetMoments ( const int width, const int height, const int N, const int nTilde,
          const bool print, const int levels )
{
    int x, y, k, n, nX, nY, step, max_step, level;
    float  maxN;
    Vector liftPtr;
    char buf[BUFSIZ];

    if (!filter) {
        Error ("GetMoments", NO_FILTER_VECTOR, ABORT);
    }

    // calculate number of iterations n. 
    maxN = (float)MAX(N, nTilde) - (float)1;   
    // x iterations
    nX = (width == 1) ? 0 : (int)logBaseN ( (float)(width-1)/maxN, (float)2 );
    nX = (nX < 0) ? 0 : MIN (levels, nX);   
    // y iterations
    nY = (height == 1) ? 0 : (int)logBaseN ( (float)(height-1)/maxN, (float)2 );
    nY = (nY < 0) ? 0 : MIN (levels, nY);  
    // total iterations
    n = MAX(nX, nY);  
    max_step = (n==0) ? 0 : 1<<n;   

    // get lifting coefficients
    level = 0;
    for ( step = 1 ; step < max_step ; step <<= 1 ) {
        if (nX-- > 0) {
            // for rows
            GetLifting (liftingX[nX], filter, step, N, nTilde, width, FALSE, false);
            if (print) {
                fprintf (stdout, "\nLifting Coefficients for Rows (Level=%d)\n",
                         level);
                liftPtr = liftingX[nX];
                for ( x=0 ; x<(width>>1) ; x++ ) {
                    sprintf (buf, "%.20g", *(liftPtr++));
                    for ( k=1 ; k<nTilde ; k++ ) {
                        sprintf (buf, "%s,%.20g", buf, *(liftPtr++));
                    }
                    fprintf (stdout, "Gamma[%d] (%s)\n", x, buf);
                }
            }
        }

        if (nY-- > 0) {
            // for columns
            GetLifting (liftingY[nY], filter, step, N, nTilde, height, TRUE, false);
            if (print) {
                fprintf (stdout,
                         "\nLifting Coefficients for Columns (Level=%d)\n",
                         level);
                liftPtr = liftingY[nY];
                for ( y=0 ; y<(height>>1) ; y++ ) {
                    sprintf (buf, "%.20g", *(liftPtr++));
                    for ( k=1 ; k<nTilde ; k++ ) {
                        sprintf (buf, "%s,%.20g", buf, *(liftPtr++));
                    }
                    fprintf (stdout, "Gamma[%d] (%s)\n", y, buf);
                }
            }
        }
        level++;  
    }
}

//------------------------------------------------
void
Lifting2DWT ( Matrix Data, const int width, const int height,
              const int N, const int nTilde,
              const int levels, const bool inverse)
{
    int i, k, x, y, n, nX, nY, step, max_step;
    float maxN;    

    // filter and lifting coefficients exit?
    if (!filter) {
        Error ("Lifting2DWT", NO_FILTER_VECTOR, ABORT);
    }

    if ( !liftingX && !liftingY ) {
        fprintf( stderr,
            "\nLifting2DWT(): lifting coefficients are not initialized."
            "\nUse GetDualWavelet() and GetMoments() before calling this function.\n");
        exit( INPUT_ERROR );
    }

    maxN = (float)MAX(N, nTilde) - (float)1;  
    // x iterations
    nX = (width == 1) ? 0 : (int)logBaseN ( (float)(width-1)/maxN, (float)2 );
    nX = (nX < 0) ? 0 : MIN(levels, nX);  
    // y iterations
    nY = (height == 1) ? 0 : (int)logBaseN ( (float)(height-1)/maxN, (float)2 );
    nY = (nY < 0) ? 0 : MIN(levels, nY);  
    n = MAX(nX, nY);  

    // forward transform
    if ( !inverse ) {  
        // get wavelet coefficients
        max_step = (n==0) ? 0 : 1<<n;  
        for ( step=1 ; step<max_step ; step<<=1 ) {
            if (nX-- > 0) {
                for ( y=0 ; y<height ; y+=step ) {
                    Lifting1DWTPredict ( &Data[y][0], (long)width, 1,
                                     (long)step, (long)N );
                    Lifting1DWTUpdate ( &Data[y][0], (long)width, 1,
                                    (long)step, (long)nTilde, liftingX[nX] );
                }
            }
    
            if (nY-- > 0) {
                for ( x=0 ; x<width ; x+=step ) {
                    Lifting1DWTPredict ( &Data[0][x], (long)height, (long)width,
                                     (long)step, (long)N );
                    Lifting1DWTUpdate ( &Data[0][x], (long)height, (long)width,
                                    (long)step, (long)nTilde, liftingY[nY] );
                }
            }
        }
    } else {  // inverse transform
        for ( i=0 ; i<N*(1+(N>>1)) ; i++ ) {
            filter[i] = -filter[i];
        }
        for ( x=0 ; x<nX ; x++ ) {
            for ( k=0 ; k<(width>>1)*nTilde ; k++ ) {
                liftingX[x][k] = -liftingX[x][k];  
            }
        }
        for ( y=0 ; y<nY ; y++ ) {
            for ( k=0 ; k<(height>>1)*nTilde ; k++ ) {
                liftingY[y][k] = -liftingY[y][k]; 
            }
        }
        max_step = (n==0) ? 0 : 1<<(n-1);   
        for ( step = max_step ; step >= 1 ; step >>= 1 ) {
            if (n <= nY) {
                for ( x=0 ; x<width ; x+=step ) {
                    Lifting1DWTUpdate ( &Data[0][x], (long)height, (long)width,
                                    (long)step, (long)nTilde, liftingY[nY-n] );
                    Lifting1DWTPredict ( &Data[0][x], (long)height, (long)width,
                                     (long)step, (long)N );
                }
            }
    
            if (n <= nX) {
                for ( y=0 ; y<height ; y+=step ) {
                    Lifting1DWTUpdate ( &Data[y][0], (long)width, 1,
                                    (long)step, (long)nTilde, liftingX[nX-n] );
                    Lifting1DWTPredict ( &Data[y][0], (long)width, 1,
                                     (long)step, (long)N );
                }
            }
            n--;   
        }
        for ( i=0 ; i<N*(1+(N>>1)) ; i++ ) {
            filter[i] = -filter[i];         
        }
        for ( x=0 ; x<nX ; x++ ) {
            for ( k=0 ; k<(width>>1)*nTilde ; k++ ) {
                liftingX[x][k] = -liftingX[x][k];   
            }
        }
        for ( y=0 ; y<nY ; y++ ) {
            for ( k=0 ; k<(height>>1)*nTilde ; k++ ) {
                liftingY[y][k] = -liftingY[y][k];  
            }
        }
    }
}

//------------------------------------------------
void
Lifting1DWT ( Vector Data, const int width,
                      const int N, const int nTilde,
                      const int levels, const bool inverse )
{
    int i, k, x, s, step, max_step, nX;            
    float maxN;          

    if (!filter) {
        Error ("Lifting1DWT", NO_FILTER_VECTOR, ABORT);
    }

    if ( !liftingX && !liftingY ) {
	fprintf( stderr,
	         "\nLifting1DWT(): lifting coefficients are not initialized."
		 "\nUse GetDualWavelet() and GetMoments() before calling this "
                 "function.\n");
        exit( INPUT_ERROR );
    }

    maxN = (float)MAX(N, nTilde) - (float)1;  
    nX = (width == 1) ? 0 : (int)logBaseN( (float)(width-1)/maxN, (float)2 );
    nX = (nX < 0) ? 0 : MIN(levels, nX);  

    // forward transform
    if ( !inverse ) {    
        max_step = (nX==0) ? 0 : 1<<nX; 
        for ( step=1, s=nX-1 ; step<max_step ; step<<=1, s-- ) {
            Lifting1DWTPredict ( Data, (long)width, 1,
                             (long)step, (long)N );
            Lifting1DWTUpdate ( Data, (long)width, 1,
                            (long)step, (long)nTilde, liftingX[s] );
        }
    } else {   // inverse transform
        for ( i=0 ; i<N*(1+(N>>1)) ; i++ )
            filter[i] = -filter[i];        
        for ( x=0 ; x<nX ; x++ )
            for ( k=0 ; k<(width>>1)*nTilde ; k++ )
                liftingX[x][k] = -liftingX[x][k];  

        max_step = (nX==0) ? 0 : 1<<(nX-1); 
        for ( step=max_step, s=0 ; step>=1 ; step>>=1, s++ ) {
            Lifting1DWTUpdate ( Data, (long)width, 1,
                            (long)step, (long)nTilde, liftingX[s] );
            Lifting1DWTPredict ( Data, (long)width, 1,
                             (long)step, (long)N );
        }

        for ( i=0 ; i<N*(1+(N>>1)) ; i++ )
            filter[i] = -filter[i];        
        for ( x=0 ; x<nX ; x++ )
            for ( k=0 ; k<(width>>1)*nTilde ; k++ )
                liftingX[x][k] = -liftingX[x][k];
    }
}


//------------------------------------------------
void
Lifting1DWTPredict ( Vector vect, const long size, const long incr,
                              const long step, const long N )
{
    register Vector lambdaPtr, gammaPtr, fP, filterPtr;  
    register long len, j, stepIncr;          
    long LlessthanR, LequalR, LgreaterthanR, soi;

    len       = CEIL(size, step);   // current level coefficients 
    stepIncr  = (step*incr) << 1;   // stepsize in between

    // get number of iterations per case
    j     = ODD(len);
    LlessthanR = N >> 1;
    LgreaterthanR = LlessthanR - j;                
    LequalR = (len >> 1) - N + 1 + j; 
    LlessthanR--;                          

    // gamma coefficients
    filterPtr = filter + N;  

    // left lamdas < right
    gammaPtr = vect + (stepIncr >> 1);  // 2nd coeff is gamma
    while(LlessthanR--) {
        lambdaPtr = vect;   // 1rst coeff is 1rst lamda
        j = N;
        do {  
            *(gammaPtr) -= (*(lambdaPtr)*(*(filterPtr++)));
            lambdaPtr   += stepIncr;  
        } while(--j);   
        gammaPtr += stepIncr;
    }

    // left lamdas equal right
    soi = 0;
    while(LequalR--) {
        lambdaPtr = vect + soi;   
        fP = filterPtr;   
        j = N;
        do {   
            *(gammaPtr) -= (*(lambdaPtr)*(*(fP++)));
            lambdaPtr   += stepIncr;   
        } while(--j);  
        soi += stepIncr;
        gammaPtr += stepIncr;
    }

    // left lamdas greater than right
    fP = filterPtr; 
    vect += (soi-stepIncr);  
    while (LgreaterthanR--) {
        lambdaPtr = vect; 
        j = N;
        do {   
            *(gammaPtr) -= (*(lambdaPtr)*(*(--fP)));
            lambdaPtr   += stepIncr;  
        } while(--j); 
        gammaPtr += stepIncr;
    }
}


//------------------------------------------------
void
Lifting1DWTUpdate ( Vector vect, const long size, const long incr,
                             const long step, const long nTilde,
                             const Vector lc )
{
    register Vector lcPtr, vL, vG;    
    register long j;       
    long len, LlessthanR, LequalR, LgreaterthanR, noGammas;
    long stepIncr, soi;           

    len      = CEIL(size, step);  
    stepIncr = (step*incr) << 1;   
    noGammas = len >> 1 ;         

    j	  = ODD(len);
    LlessthanR = nTilde >> 1;
    LgreaterthanR = LlessthanR - j;                 
    LequalR = noGammas - nTilde + 1 + j;   
    LlessthanR--;                             

    // lift lamdas
    lcPtr = lc;  

    // left lambdas less than right
    vG = vect + (stepIncr >> 1);   
    while(LlessthanR--) {
        vL = vect; 
        j = nTilde;
        do {
            *(vL) += (*(vG)*(*(lcPtr++))); 
            vL    += stepIncr;           
        } while(--j);   
        vG += stepIncr;
    }

    // equal lambdas
    soi = 0;
    while(LequalR--) {
        vL = vect + soi; 
        j = nTilde;
        do {
            *(vL) += (*(vG)*(*(lcPtr++))); 
            vL    += stepIncr;             
        } while(--j);  
        vG  += stepIncr;
        soi += stepIncr;
    }

    // right more than left
    vect += (soi - stepIncr);
    while(LgreaterthanR--) {
        vL = vect; 
        j = nTilde;
        do {
            *(vL) += (*(vG)*(*(lcPtr++)));  
            vL    += stepIncr;             
        } while(--j);   
        vG += stepIncr;
    }
}

//------------------------------------------------
void
HaarLifting2DWT ( Matrix Data, const int width, const int height,
                          const int levels, const bool inverse )
{
    int x, y, n, nX, nY, step, max_step;
                   
    // number of iterations
    nX = (width==1) ? 0 : (int)ceil((float)logBaseN((float)(width),2.0));
    nX = (nX<0) ? 0 : MIN(nX,levels);
    nY = (height==1) ? 0 : (int)ceil((float)logBaseN((float)(height),2.0));
    nY = (nX<0) ? 0 : MIN(nY,levels);
    n = MAX(nX,nY);

    // forward transform
    if ( !inverse ) {   
        // calc wavelet coefficients
        max_step = (n==0) ? 0 : 1<<n;   // max step
        for ( step = 1 ; step < max_step ; step <<= 1 ) {
            if (nX-- > 0) {
                for ( y=0 ; y<height ; y+=step ) {
                    HaarLambdaCoeffs ( &Data[y][0], (long)width, 1, (long)step );
                }
            }
    
            if (nY-- > 0) {
                for ( x=0 ; x<width ; x+=step ) {
                    HaarLambdaCoeffs ( &Data[0][x], (long)height, (long)width,
                                  (long)step );
                }
            }
        }
    } else {   // Inverse transform 
        max_step = (n==0) ? 0 : 1<<(n-1);  
        for ( step = max_step ; step >= 1 ; step >>= 1 ) {
            if (n <= nY) {
                for ( x=0 ; x<width ; x+=step ) {
                    InvHaarLambdaCoeffs ( &Data[0][x], (long)height, (long)width,
                                   (long)step );
                }
            }
    
            if (n <= nX) {
                for ( y=0 ; y<height ; y+=step ) {
                    InvHaarLambdaCoeffs ( &Data[y][0], (long)width, 1, (long)step );
                }
            }
            n--;  
        }
    }
}

//------------------------------------------------
void
HaarLifting1DWT ( Vector Data, const int width,
                           const int levels, const bool inverse )
{
    int s, nX;       

    nX = (width==1) ? 0 :
        (int)ceil((float)logBaseN((float)(width),2.0));
    nX = (nX<0) ? 0 : MIN(nX,levels);

    // forward transform
    if ( !inverse ) {   
        for ( s = 0 ; s < nX ; s++ ) {
            HaarLambdaCoeffs ( Data, (long)width, 1, (long)(1<<s) );
        }
    } else {  // inverse
        for ( s = nX-1 ; s >=0 ; s-- ) {
            InvHaarLambdaCoeffs ( Data, (long)width, 1, (long)(1<<s) ); 
        }
    }
}

//------------------------------------------------
void
HaarLambdaCoeffs ( Vector vect, const long size, const long incr,
                         const long step )
{
    register Vector lambdaPtr, gammaPtr; 
    register long nL, nG, stepIncr;
    long len; 
    float  lAve, lSum = 0;


    len      = CEIL(size, step);  
    stepIncr = (step*incr) << 1; 
    nG = (len >> 1);
    nL = len - nG;

    // calc lambda average
    if ( ODD(len) ) {
        long n = len, s = stepIncr >> 1;
        float  t = 0;
        float *l = vect;

        while (n--) {
            t += *l;
            l += s;
        }
        lAve = t/(float)len;  
    }

    gammaPtr = vect + (stepIncr >> 1);   
    lambdaPtr = vect;                    
    while (nG--) {
        *(gammaPtr) -= *(lambdaPtr);
        *(lambdaPtr) += (*(gammaPtr)/(float)2);
        lSum += *(lambdaPtr);

        gammaPtr += stepIncr;
        lambdaPtr += stepIncr;
    }
    if ( ODD(len) ) {
        *(lambdaPtr) = lAve*(float)nL - lSum;
        lSum += *(lambdaPtr);
    }
}

//------------------------------------------------
void
InvHaarLambdaCoeffs( Vector vect, const long size, const long incr,
                         const long step )
{
    register Vector lambdaPtr, gammaPtr;
    register long nL, nG, stepIncr;     
    long len;                
    float  lAve, lSum = 0;            

    len      = CEIL(size, step);  
    stepIncr = (step*incr) << 1;   

    nG = (len >> 1);
    nL = len - nG;

    if ( ODD(len) ) {
        long n = nL, s = stepIncr;
        float  t = 0;
        float *l = vect;

        while (n--) {
            t += *l;
            l += s;
        }
        lAve = t/(float)nL;   /* average we've to maintain */
    }

    gammaPtr = vect + (stepIncr >> 1);
    lambdaPtr = vect;                   
    while (nG--) {
        *(lambdaPtr) -= (*(gammaPtr)/(float)2);
        lSum += *(lambdaPtr);
    	*(gammaPtr) += *(lambdaPtr);
        lSum += *(gammaPtr);
        gammaPtr += stepIncr;
        lambdaPtr += stepIncr;
    }
    if ( ODD(len) ) {
        *(lambdaPtr) = lAve*(float)len - lSum;
        lSum += *(lambdaPtr);
    }
}

//------------------------------------------------
void
LiftingToMallat ( Matrix Data, const int width, const int height,
                                const int N, const int nTilde,
                                const int levels)
{
    int i, x, y, lenX, lenX2, lenY, lenY2, n, nX, nY;   
    float maxN;   

    if ( N==1 || nTilde==1 ) {   // Haar wavelets
        nX = (width==1) ? 0 :(int)ceil((float)logBaseN((float)width,2.0));
        nY = (height==1) ? 0 : (int)ceil((float)logBaseN((float)height,2.0));
    } else {                     // Biorthogonal wavelets
        maxN = (float)MAX(N, nTilde) - (float)1;
        nX = (width==1) ? 0 : (int)logBaseN((float)(width-1)/maxN,2.0);
        nY = (height==1) ? 0 :(int)logBaseN((float)(height-1)/maxN,2.0);
    }
    nX = (nX < 0) ? 0 : MIN (levels, nX);
    nY = (nY < 0) ? 0 : MIN (levels, nY);
    n  = MAX(nX, nY);

    lenX = lenX2 = width;
    lenY = lenY2 = height;

    // Apply Mallat format
    for ( i=0 ; i<n ; i++ ) {
        if (nX-- > 0) {
            for ( y=0 ; y<lenY2 ; y++ ) {
                LiftingToMallat1D( &Data[y][0], lenX, 1);
            }
            lenX  -= (lenX  >> 1);
	        lenY2 -= (lenY2 >> 1);
        }
        if (nY-- > 0) {
            for ( x=0 ; x<lenX2 ; x++ ) {
                LiftingToMallat1D( &Data[0][x], lenY, width);
            }
            lenY  -= (lenY  >> 1);
            lenX2 -= (lenX2 >> 1);
        }
    }
}

//------------------------------------------------
void
MallatToLifting ( Matrix Data, const int width, const int height,
                                const int N, const int nTilde,
                                const int levels)
{
    int i, x, y, lenX, lenX2, lenY, lenY2, n, nX, nY;   
    float maxN;   

    if ( N==1 || nTilde==1 ) {   // Haar wavelets
        nX = (width==1) ? 0 :(int)ceil((float)logBaseN((float)width,2.0));
        nY = (height==1) ? 0 : (int)ceil((float)logBaseN((float)height,2.0));
    } else {                     // Biorthogonal wavelets
        maxN = (float)MAX(N, nTilde) - (float)1;
        nX = (width==1) ? 0 : (int)logBaseN((float)(width-1)/maxN,2.0);
        nY = (height==1) ? 0 :(int)logBaseN((float)(height-1)/maxN,2.0);
    }
    nX = (nX < 0) ? 0 : MIN (levels, nX);
    nY = (nY < 0) ? 0 : MIN (levels, nY);
    n  = MAX(nX, nY);

    for ( i=n-1 ; i>=0 ; i-- ) {
        int j;
        lenX = lenX2 = width;
        lenY = lenY2 = height;
        if (nY-- > 0) {
            for ( j=0 ; j<i ; j++ ) {
                lenY  -= (lenY  >> 1);
                lenX2 -= (lenX2 >> 1);
            }
            for ( x=0 ; x<lenX2 ; x++ ) {
                MallatToLifting1D( &Data[0][x], lenY, width);
            }
        }
        if (nX-- > 0) {
            for ( j=0 ; j<i ; j++ ) {
                lenX  -= (lenX  >> 1);
                lenY2 -= (lenY2 >> 1);
            }
            for ( y=0 ; y<lenY2 ; y++ ) {
                MallatToLifting1D( &Data[y][0], lenX, 1);
            }
        }
    }
}

//------------------------------------------------
void LiftingToMallat1D ( Vector Data, const int len, const int offset)
{
    register int i;
    int noLambdas, noGammas;
    Vector lPtr, gPtr, bufPtr;
    Vector buffer;

    buffer = AllocFloatVector( 0, (long)(len-1) );
    noGammas  = len>>1;
    noLambdas = len - noGammas;
    
    lPtr = buffer;           
    gPtr = buffer + (noLambdas); 

    // lambdas to the left and all gammas to the right
    for ( i=0 ; i<len-2 ; i+=2 ) {
        *(lPtr) = Data[i*offset];
        *(gPtr) = Data[(i+1)*offset];
        lPtr ++;
        gPtr ++;
    }
    if ( ODD(len) ) {
        i = len-1;
        *(lPtr) = Data[i*offset];
    } else {
        i = len-2;
        *(lPtr) = Data[i*offset];
        *(gPtr) = Data[(i+1)*offset];
    }
    // results into original vector
    bufPtr = buffer;
    for ( i=0 ; i<len*offset ; i+=offset ) {
	Data[i] = *(bufPtr++);
    }

    FreeFloatVector( buffer, 0, (long)(len-1) );
}

//------------------------------------------------
void MallatToLifting1D ( Vector Data, const int len, const int offset)
{
    register int i;
    int noLambdas, noGammas;
    Vector lPtr, gPtr, bufPtr;
    Vector buffer;

    buffer = AllocFloatVector( 0, (long)(len-1) );
    noGammas  = len>>1;
    noLambdas = len - noGammas;
    
    lPtr = Data;                   
    gPtr = Data + (noLambdas)*offset;   

    // Alternate lambdas and gammas
    for ( i=0 ; i<len-2 ; i+=2 ) {
        buffer[i]   = *(lPtr);
        buffer[i+1] = *(gPtr);
        lPtr += offset;
        gPtr += offset;
    }
    if ( ODD(len) ) {
        i = len-1;
        buffer[i] = *(lPtr);
    } else {
        i = len-2;
        buffer[i]   = *(lPtr);
        buffer[i+1] = *(gPtr);
    }
    // results into original vector
    bufPtr = buffer;
    for ( i=0 ; i<len*offset ; i+=offset ) {
	Data[i] = *(bufPtr++);
    }

    FreeFloatVector( buffer, 0, (long)(len-1) );
}

//------------------------------------------------
static void
PrintMatrix ( const Matrix A, const int sizeX, const int sizeY )
{
    int i, j;

    for ( i=0 ; i<sizeY ; i++ ) {
        fprintf (stdout, "[");
        for ( j=0 ; j<sizeX-1 ; j++ ) {
            fprintf (stdout, "%3.2f, ", A[i][j]);
        }
        fprintf (stdout, "%3.2f]\n", A[i][j]);
    }
}

//------------------------------------------------
static void
Scale ( Matrix A, const int width, const int height,
                  const int Min  , const int Max  , const int gamma )
{
    int i, j;
    float  min, max, minT, maxT;

    MinMax (A, width, height, &min, &max);
    fprintf (stdout, "Min(%4.2f) Max(%4.2f)\t", min, max);
    if (min < (float)Min) {
        for ( i=0 ; i<height ; i++ ) {
            for ( j=0 ; j<width ; j++ ) {
                A[i][j] += (-min); 
            }
        }
        MinMax (A, width, height, &minT, &maxT);
        fprintf (stdout, "\n\t    Min (%4.2f) Max (%4.2f)\t", minT, maxT);
    }
    if ( max>(float)Max || gamma>1 ) {
        if ( gamma>1 ) fprintf (stdout, "\n\t    Gamma (%d)\t", gamma);
        else fprintf (stdout, "\n\t    (No gamma correction)\t");
        fflush (stdout);
        for ( i=0 ; i<height ; i++ ) {
            for ( j=0 ; j<width ; j++ ) {
                A[i][j] = A[i][j]/(max-min); 
                if ( gamma>1 && A[i][j]!=(float)0 ) {  
                    A[i][j] = (float)pow((float)A[i][j], (float)1/(float)gamma);
                }
                A[i][j] *= (float)255;           
            }
        }
    }
    if ( gamma>1 ) {
        MinMax (A, width, height, &min, &max);
        fprintf (stdout, "\n\t    Min (%4.2f) Max (%4.2f)\t", min, max);
    } 
}

//------------------------------------------------
static void
MinMax ( const Matrix A, const int width, const int height, float *min, float *max )
{
    int i, j;

    *min = FLT_MAX;
    *max = FLT_MIN;

    for ( i=0 ; i<height ; i++ ) {
        for ( j=0 ; j<width ; j++ ) {
            *min = MIN(*min, A[i][j]);
            *max = MAX(*max, A[i][j]);
        }
    }
}
